import time
import random
import logging

from selenium.common.exceptions import TimeoutException, WebDriverException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC

from .utils import setup_firefox_driver, try_to_click
from .exceptions import IncorrectUsernameOrPasswordException

PAGE_LOAD_TIMEOUT_S = 60
DRIVER_TIMEOUT_S = 60
MICRO_DELAY_S = 1

# Random subreddits to subscribe to
subs = [
    'gaming',
    'csgo',
    'memes',
    'tf2',
    'askreddit',
    'abruptchaos',
    'perfectlycutscreams',
    'askouija',
    'apple',
    'linux',
    'pcmasterrace',
    'clashroyale',
    'minecraft',
    'cursed_comments',
    'shitposting',
]
_logger = logging.getLogger(__name__)


def protect_account(username: str, password: str,
                    proxies: dict[str, str] | None = None, hide_browser: bool = True):
    """
    Prevent account blocking due to suspicion of being a bot.

    This function will:
    - Subscribe to random subreddit
    - For now, that's all
    """

    logging.info('Protecting account with username %s', username)
    driver = setup_firefox_driver(proxies, hide_browser)

    if PAGE_LOAD_TIMEOUT_S is not None:
        driver.set_page_load_timeout(PAGE_LOAD_TIMEOUT_S)

    try:  # try/except to quit driver if error occurs

        # Open login website
        _logger.debug('Opening login page')
        try:
            driver.get('https://www.reddit.com/login/')
        except WebDriverException:
            raise TimeoutException('Website takes too long to load. Probably a problem with the proxy.')
        
        # Enter username and password
        _logger.debug('Entering username and password')
        username_input = driver.find_element(By.ID, 'loginUsername')
        password_input = driver.find_element(By.ID, 'loginPassword')
        try_to_click(username_input, delay=MICRO_DELAY_S)
        username_input.send_keys(username)
        try_to_click(password_input, delay=MICRO_DELAY_S)
        password_input.send_keys(password)

        # Submit login
        _logger.debug('Submitting login')
        submit_btn = driver.find_element(By.CSS_SELECTOR, 'button[type="submit"]')
        WebDriverWait(driver, DRIVER_TIMEOUT_S).until(EC.element_to_be_clickable(submit_btn))
        try_to_click(submit_btn, delay=MICRO_DELAY_S)

        # Check if submitted
        time.sleep(MICRO_DELAY_S)
        submit_err = driver.find_element(By.CLASS_NAME, 'AnimatedForm__errorMessage')

        if submit_err.text != '':
            if 'incorrect' in submit_err.text.lower():
                raise IncorrectUsernameOrPasswordException(submit_err.text)
            raise Exception(submit_err.text)

        # Logged in!

        # Go to random subreddit
        sub = random.choice(subs)
        _logger.debug('Going to subreddit r/%s', sub)
        WebDriverWait(driver, DRIVER_TIMEOUT_S).until(EC.url_matches('https://www.reddit.com*'))
        driver.get(f'https://www.reddit.com/r/{sub}/')

        # Subscribe
        _logger.debug('Subscribing to subreddit r/%s', sub)
        subscribe_btn = driver.find_element(By.XPATH, '//button[text()="Join"]')
        WebDriverWait(driver, DRIVER_TIMEOUT_S).until(EC.element_to_be_clickable(subscribe_btn))
        try_to_click(subscribe_btn, delay=MICRO_DELAY_S)

        # Done!

    except Exception as e:  # quit driver if error occurs
        driver.quit()
        raise e

    driver.quit()



import time
from abc import ABC, abstractmethod

from stem import Signal
from stem.control import Controller


class ProxyManager(ABC):
    """
    Abstract class for proxy managers

    Proxy manager needed to switch proxies after each account creation
    """

    @abstractmethod
    def get_next(self) -> dict[str, str] | None:
        ...


class DefaultProxy(ProxyManager):
    """Proxy manager based on proxies list"""

    def __init__(self, proxies: list[str]):
        self.proxies = proxies
        self.i = 0

    def get_next(self):
        proxy = {
            'socks': self.proxies[self.i]
        }

        self.i += 1
        if self.i >= len(self.proxies):
            self.i = 0

        return proxy

    def __str__(self) -> str:
        return self.proxies[self.i]


class TorProxy(ProxyManager):
    """Proxy manager based on Tor"""

    def __init__(self, ip: str, port: int, password: str, control_port: int, delay: int = 5):
        self.ip = ip
        self.port = port
        self.password = password
        self.control_port = control_port
        self.delay = delay

        self.controller = Controller.from_port(port=self.control_port)
        self.controller.authenticate(self.password)

    @property
    def proxy(self) -> dict[str, str]:
        return {
            'http': f'{self.ip}:{self.port}',
            'https': f'{self.ip}:{self.port}',
        }

    def get_next(self):
        self.controller.signal(Signal.NEWNYM)
        time.sleep(self.delay)

        return self.proxy

    def __str__(self) -> str:
        return f'{self.ip}:{self.port}'


class EmptyProxy(ProxyManager):
    """In case if we want to use local IP address"""
    def get_next(self):
        return None

    def __str__(self) -> str:
        return 'No proxy'



import os
import time
import random
import string
import secrets

import requests
from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.remote.webelement import WebElement
from random_username.generate import generate_username as _generate_username


def generate_username() -> str:
    username = _generate_username(1)[0] + str(random.randint(100, 1000))
    return username


def generate_password(length: int = 12) -> str:
    characters = string.ascii_letters + string.digits
    password = ''.join(secrets.choice(characters) for _ in range(length))
    return password


def load_proxies(path: str) -> list[str]:
    """
    Load proxies from file

    File format:
    IP:PORT without protocol
    """

    if not os.path.exists(path):
        return []

    proxies = []

    with open(path, 'r', encoding='utf-8') as f:
        for _, line in enumerate(f):
            line = line.strip()
            if line == '' or line.startswith('#'):
                continue
            proxies.append(line)

    return proxies


def check_tor_running(ip: str, port: int) -> bool:
    try:
        r = requests.get('https://check.torproject.org/api/ip', proxies={'https': f'socks5h://{ip}:{port}'}, timeout=5)
        return r.json()['IsTor'] is True
    except Exception:
        return False


def setup_chrome_driver(proxies: dict[str, str] | None = None, hide_browser: bool = True) -> webdriver.Chrome:
    options = webdriver.ChromeOptions()
    options.add_argument('--lang=en-US')

    if hide_browser:
        options.add_argument('--headless')

    # Set up proxies if available
    if proxies is not None:
        if 'http' in proxies:
            options.add_argument(f'--proxy-server=http://{proxies["http"]}')
        if 'https' in proxies:
            options.add_argument(f'--proxy-server=https://{proxies["https"]}')
        if 'socks' in proxies:
            # Only SOCKS5 is supported
            options.add_argument(f'--proxy-server=socks5://{proxies["socks"]}')

    return webdriver.Chrome(options=options, service_log_path=os.devnull)


def setup_firefox_driver(proxies: dict[str, str] | None = None, hide_browser: bool = True) -> webdriver.Firefox:
    options = webdriver.FirefoxOptions()
    options.set_preference('intl.accept_languages', 'en-US')

    if hide_browser:
        options.add_argument('--headless')

    # Set up proxies if available
    if proxies is not None:
        options.set_preference('network.proxy.type', 1)

        if 'http' in proxies:
            http_ip, http_port = proxies['http'].split(':')
            options.set_preference('network.proxy.http', http_ip)
            options.set_preference('network.proxy.http_port', int(http_port))
        if 'https' in proxies:
            https_ip, https_port = proxies['https'].split(':')
            options.set_preference('network.proxy.ssl', https_ip)
            options.set_preference('network.proxy.ssl_port', int(https_port))
        if 'socks' in proxies:
            # Only SOCKS5 is supported
            socks_ip, socks_port = proxies['socks'].split(':')
            options.set_preference('network.proxy.socks', socks_ip)
            options.set_preference('network.proxy.socks_port', int(socks_port))
            options.set_preference('network.proxy.socks_remote_dns', False)

    return webdriver.Firefox(options=options, service_log_path=os.devnull)


def try_to_click(element: WebElement, delay: int | float = 0.5, max_tries: int = 20) -> bool:
    """Try to click an element multiple times."""
    retries = 0
    while retries < max_tries:
        try:
            element.click()
            return
        except:
            retries += 1
            time.sleep(delay)
    raise TimeoutException(f'Could not click element after {max_tries} tries.')



import time
import logging

from selenium.common.exceptions import TimeoutException, WebDriverException, NoSuchElementException, ElementClickInterceptedException
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium_recaptcha_solver import RecaptchaSolver, RecaptchaException

from .utils import setup_firefox_driver, try_to_click, generate_password, generate_username
from .exceptions import *

PAGE_LOAD_TIMEOUT_S = 60
DRIVER_TIMEOUT_S = 60
MICRO_DELAY_S = 1

_logger = logging.getLogger(__name__)


def create_account(email: str, username: str | None = None, password: str | None = None,
                   proxies: dict[str, str] | None = None, hide_browser: bool = True) -> tuple[str, str]:
    """Create a Reddit account."""

    _logger.info('Creating reddit account')
    driver = setup_firefox_driver(proxies, hide_browser)

    if PAGE_LOAD_TIMEOUT_S is not None:
        driver.set_page_load_timeout(PAGE_LOAD_TIMEOUT_S)

    if password is None:
        password = generate_password()

    try:  # try/except to quit driver if error occurs

        # Open website
        _logger.debug('Opening registration page')
        try:
            driver.get('https://www.reddit.com/register/')
        except WebDriverException:
            raise TimeoutException('Website takes too long to load. Probably a problem with the proxy.')

        # Checking if IP is blocked
        try:
            first_h1 = driver.find_element(By.TAG_NAME, 'h1')
            if first_h1.text == 'whoa there, pardner!':
                raise IPCooldownException('Your IP is temporarily blocked. Try again later.')
        except NoSuchElementException:
            pass

        # Enter email and go to next page
        _logger.debug('Entering email')
        email_input = driver.find_element(By.ID, 'regEmail')
        email_submit = driver.find_element(By.CSS_SELECTOR, 'button[data-step="email"]')
        email_input.click()
        email_input.send_keys(email)
        try_to_click(email_submit, delay=MICRO_DELAY_S)

        # Check for email error
        time.sleep(MICRO_DELAY_S)
        try:
            email_err = driver.find_element(By.CLASS_NAME, 'AnimatedForm__errorMessage')
        except:
            pass
        else:
            if email_err.text != '':
                if 'again' in email_err.text.lower():
                    raise SessionExpiredException(email_err.text)
                raise Exception(email_err.text)

        # Wait until page loads
        WebDriverWait(driver, DRIVER_TIMEOUT_S).until(EC.element_to_be_clickable((By.ID, 'regUsername')))

        # Enter username
        _logger.debug('Entering username and password')
        try:
            # Get random username suggested by reddit
            random_username = driver.find_element(By.XPATH, '/html/body/div/main/div[2]/div/div/div[2]/div[2]/div/div/a[1]')
        except NoSuchElementException:
            # Sometimes reddit doesn't suggest any username
            username = generate_username()

        username_input = driver.find_element(By.ID, 'regUsername')
        if username is not None:
            # Enter given username
            try_to_click(username_input, delay=MICRO_DELAY_S)
            username_input.send_keys(username)
        else:
            # Click first reddit sugegsted name
            try_to_click(random_username, delay=MICRO_DELAY_S)
            username = random_username.text

        # Enter password
        password_input = driver.find_element(By.ID, 'regPassword')
        try_to_click(password_input, delay=MICRO_DELAY_S)
        password_input.send_keys(password)

        # Check for username and password errors
        username_err = driver.find_element(By.XPATH, '//div[@data-for="username"]')
        password_err = driver.find_element(By.XPATH, '//div[@data-for="password"]')
        try_to_click(username_input, delay=MICRO_DELAY_S)
        time.sleep(MICRO_DELAY_S)

        if username_err.text != '':
            if 'taken' in username_err.text.lower():
                raise UsernameTakenException(username_err.text)
            if 'character' in username_err.text.lower():
                raise UsernameLengthException(username_err.text)
            if 'symbols' in username_err.text.lower():
                raise UsernameSymbolsException(username_err.text)
            raise Exception(username_err.text)

        if password_err.text != '':
            if 'character' in password_err.text.lower():
                raise PasswordLengthException(password_err.text)
            raise Exception(password_err.text)
        
        # Solve captcha
        _logger.debug('Solving captcha')
        WebDriverWait(driver, DRIVER_TIMEOUT_S).until(EC.element_to_be_clickable((By.XPATH, '//iframe[@title="reCAPTCHA"]')))
        recaptcha_iframe = driver.find_element(By.XPATH, '//iframe[@title="reCAPTCHA"]')

        if recaptcha_iframe.is_displayed():
            solver = RecaptchaSolver(driver)
            for _ in range(5):
                try:
                    solver.click_recaptcha_v2(iframe=recaptcha_iframe)
                    break
                except ElementClickInterceptedException:
                    pass
            else:
                raise RecaptchaException('Could not solve captcha')

        # Submit registration
        _logger.debug('Submitting registration')
        submit_btn = driver.find_element(By.CSS_SELECTOR, 'button[data-step="username-and-password"]')
        WebDriverWait(driver, DRIVER_TIMEOUT_S).until(EC.element_to_be_clickable(submit_btn))
        try_to_click(submit_btn, delay=MICRO_DELAY_S)

        # Check if submitted
        time.sleep(MICRO_DELAY_S)
        submit_err = driver.find_element(By.CSS_SELECTOR, 'span[data-step="username-and-password"]')

        if submit_err.text != '':
            if 'again' in submit_err.text.lower():
                raise IPCooldownException(submit_err.text)
            raise Exception(submit_err.text)

        # Wait until button is pressed
        time.sleep(MICRO_DELAY_S)

        # Account created!

    except Exception as e:  # quit driver if error occurs
        driver.quit()
        raise e

    driver.quit()
    return username, password



from selenium.common.exceptions import TimeoutException
from selenium_recaptcha_solver.exceptions import RecaptchaException


class UsernameTakenException(Exception):
    pass

class UsernameLengthException(Exception):
    pass

class UsernameSymbolsException(Exception):
    pass

class PasswordLengthException(Exception):
    pass

class IPCooldownException(Exception):
    pass

class SessionExpiredException(Exception):
    pass

class IncorrectUsernameOrPasswordException(Exception):
    pass

NetworkException = (RecaptchaException, IPCooldownException, TimeoutException)



import re
from setuptools import setup, find_packages


def read(path: str) -> str:
    with open(path, 'r', encoding='utf-8') as f:
        return f.read()


def get_version(ver_file: str) -> str:  # Credits to pyTelegramBotAPI setup.py
    with open(ver_file, 'r', encoding='utf-8') as f:
        return re.search(r"^__version__\s*=\s*'(.*)'.*$",
                         f.read(), flags=re.MULTILINE).group(1)


setup(
    name='reddit-account-generator',
    version=get_version('reddit_account_generator/_version.py'),
    description='Automatic reddit account generator on selenium.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='cubicbyte',
    author_email='bmaruhnenko@gmail.com',
    url='https://github.com/cubicbyte/reddit-account-generator',
    packages = find_packages(),
    license='MIT',
    keywords='python reddit account-generator account-maker account-generation r-place rplace',
    install_requires=[
        'selenium >=4.7.0, <=4.9.1',
        'selenium-recaptcha-solver >=1.4.0, <=1.9.0',
        'random-username >=1.0.0, <=1.0.2',
        'webdriverdownloader >=1.0.0, <=1.1.0.3',
        'stem >=1.8.0, <=1.8.2',
        'static-ffmpeg >=2.3, <=2.5',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
    ],
)



# Download necessary ffmpeg binary
import static_ffmpeg
static_ffmpeg.add_paths()
del static_ffmpeg

from .maker import create_account
from .protector import protect_account


def install_driver():
    """Install firefox driver binary."""
    import webdriverdownloader
    try:
        webdriverdownloader.GeckoDriverDownloader().download_and_install()
    except RuntimeError:
        raise RuntimeError('Failed to install firefox driver. You\'re probably running the application too many times. Try again later. Limit: 60 requests per hour')
